/*******************************************************************************
 *	BNetworkHttp.h		BNetworkHttp class
 *			T.Barnaby,	BEAM Ltd,	2010-06-28
 *******************************************************************************
 */
#ifndef BHttp_H
#define BHttp_H	1

#include <stdint.h>
#include <BError.h>
#include <BSocket.h>
#include <BThread.h>
#include <BDict.h>
#include <BFifo.h>
#include <BHtmlPage.h>
#include <BEvent.h>

#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#ifndef DBG_THREADS
#define DBG_THREADS		0x0100
#endif
#define DBG_HTTP		0x0200

enum HttpResponseNum {
	HTTP_CONTINUE = 100,
	HTTP_SWITCHING_PROTOCOLS = 101,
	HTTP_OK = 200,
	HTTP_CREATED = 201,
	HTTP_ACCEPTED = 202,
	HTTP_NON_AUTHORITATIVE_INFO = 203,
	HTTP_NO_CONTENT = 204,
	HTTP_MULTIPLE_CHOICES = 300,
	HTTP_MOVED_PERMANENTLY = 301,
	HTTP_MOVED_TEMPORARILY = 302,
	HTTP_NOT_MODIFIED = 304,
	HTTP_BAD_REQUEST = 400,       /* malformed syntax */
	HTTP_UNAUTHORIZED = 401, /* authentication needed, respond with auth hdr */
	HTTP_PAYMENT_REQUIRED = 402,
	HTTP_FORBIDDEN = 403,
	HTTP_NOT_FOUND = 404,
	HTTP_REQUEST_TIMEOUT = 408,
	HTTP_INTERNAL_SERVER_ERROR = 500,
	HTTP_NOT_IMPLEMENTED = 501,   /* used for unrecognized requests */
	HTTP_BAD_GATEWAY = 502,
	HTTP_SERVICE_UNAVAILABLE = 503, /* overload, maintenance */
	HTTP_RESPONSE_SETSIZE = 0xffffffff
};

class BHttpHeader : public BDictString {
public:
	void	dump();
};

class BHttpConn {
public:
		BHttpConn(int fd, SSL* ssl = 0);
	
	int	fd;
	SSL*	ssl;
};

class BHttpServer;

class BHttpConnection: public BThread {
public:
	enum			{ BufSize = 8192 };
	enum FormType		{ FormTypeStd, FormTypeMulti }; 

				BHttpConnection(BHttpServer& server, BHttpConn conn, BSocketAddressINET from);
	virtual			~BHttpConnection();

	void			start();
	BHttpConn		getConn();

protected:
	virtual BError		process();
	virtual BError		processFile();
	virtual BError		processInternal();

	BError			redirect(BString url);
	
	BError			recv(void* buf, BUInt32 nBytes, BUInt32& nTrans);
	BError			recvLineGet(BUInt32& num, int waitForLine);
	BError			recvUpToLine(void* buf, BUInt32 nBytes, BUInt32& nTrans);
	BError			recvLine(BString& line);
	BError			recvHeader(BHttpHeader& header);

	BError			send(const void* buf, BUInt32 nBytes, BUInt32& nTrans);
	BError			sendString(BString line);
	BError			sendHeader(int num, BString contentType = "text/plain", BUInt contentLength = 0, int cache = 0, int chunked = 0, BStringList extra = BStringList());
	BError			sendHttpError(int num, BString str = "", BString info = "");

	BString			encodeString(BString str);
	BString			decodeString(BString str, int option);
	BString			htmlStrip(BString str);

	void			cookieSet(BString name, BString value);
	BString			cookieGet(BString name);

protected:
	BHttpServer&		oserver;
	BHttpConn		oconn;
	BFifo<char>		orecvFifo;
	BHttpHeader		oheader;
	BString			ocmd;
	BString			ourl;
	BString			opath;
	BDictString		oget;
	BDictString		opost;
	BDictString		ocookies;
	BDictString		ocookiesSend;
	
private:
	void*			function();
	BSocketAddressINET	ofrom;
};

// Simple HTTP server
class BHttpServer: public BThread {
public:
				BHttpServer();
	virtual			~BHttpServer();

	// Main control functions		
	BError			init(BString htmlDir, int port = 80);
	void			clientGone(BHttpConnection* client);

	// Server state
	BString			name();
	BError			setName(BString name);
	BString			htmlDir();
	BString			uploadDir();
	void			setUploadDir(BString uploadDir);
	BStringList&		internalFunctions();
	BError			setInternalFunctions(BStringList internalFunctions);

	// Main override functions
	virtual BHttpConnection*	newConnection(BHttpConn conn, BSocketAddressINET from);
	virtual BHtmlPage*		newPage();
	virtual BError			recv(BHttpConn conn, void* data, BUInt32 nBytes, BUInt32& nTrans);
	virtual BError			send(BHttpConn conn, const void* data, BUInt32 nBytes, BUInt32& nTrans);
	virtual void			closeConnection(BHttpConn conn);
	
	// Misc
	void			setDebug(int level);
	int			debug();

protected:
	virtual void*		function();				// Thread execution
	
	int			odebug;
	BString			oname;
	BString			ohtmlDir;
	BString			ouploadDir;
	BSocket			osocket;
	BList<BHttpConnection*>	oconnections;
	BEventPipe		oclientGoneEvent;
	BStringList		ointernalFunctions;
};

// Simple TLS/SSL secure HTTP server
class BHttpServerSecure: public BHttpServer {
public:
				BHttpServerSecure();
	virtual			~BHttpServerSecure();

	// Main control functions		
	BError			init(BString htmlDir, int port, BString keyFile, BString certFile);

	// Main override functions
	virtual BHttpConnection*	newConnection(BHttpConn conn, BSocketAddressINET from);
	virtual BError			recv(BHttpConn conn, void* data, BUInt32 nBytes, BUInt32& nTrans);
	virtual BError			send(BHttpConn conn, const void* data, BUInt32 nBytes, BUInt32& nTrans);
	virtual void			closeConnection(BHttpConn conn);
	
private:
	virtual void*		function();				// Thread execution

	SSL_CTX*		osslContext;
};

#endif
