/*******************************************************************************
 *	BTimeUs.h	BTimeUs functions
 *	T.Barnaby,	Beam Ltd,	2018-08-15
 *	Copyright (c) 2018 All Right Reserved, Beam Ltd, http://www.beam.ltd.uk
 *******************************************************************************
 *
 * Implements a microsecond accurate time class. Stores the time as a number of microseconds
 *  to the TAI standard with no leap seconds. 0 time is 1970-01-02T00:00:00.
 * The class provides conversions between UTC time and TAI microsecond time (leap second handling)
 *
 * Uses some sepcial values.
 *  0 - DateTime not set.
 */
#ifndef BTimeUs_h
#define BTimeUs_h

#include <BTypes.h>
#include <BError.h>
#include <BString.h>
#include <BTime.h>

class BTimeUs {
public:
			BTimeUs(BUInt64 t = 0);
			BTimeUs(BTime t);
	
	void		set(BUInt64 microSeconds);										///< Set the time to TAI us
	void		set(BUInt year, BUInt month, BUInt day, BUInt hour = 0, BUInt minute = 0, BUInt second = 0, BUInt microSecond = 0);	///< Set the date and time from UTC
	void		setYearDay(BUInt year, BUInt yearDay, BUInt hour = 0, BUInt minute = 0, BUInt second = 0, BUInt microSecond = 0);	///< Set the date and time from UTC
	
	void		getDate(BUInt& year, BUInt& month, BUInt& day) const;		///< Return the date information UTC
	void		getTime(BUInt& hour, BUInt& minute, BUInt& second) const;	///< Return the time information UTC
	BUInt64		getSeconds() const;						///< Return the number of seconds TAI
	BUInt64		getMicroSeconds() const;					///< Return the number of micro seconds TAI

	int		isSet() const{ return otime != 0; }			///< Check if set
	int		isLeapYear();						///< Returns if a leap year
	void		addSeconds(BInt64 seconds);				///< Add the given number of seconds.
	void		addMicroSeconds(BInt64 microSeconds);			///< Add the given number of seconds.

	BString		getString(BString format = "isoT") const;		///< Gets the date/time in string format
	BString		getStringUs(BString format = "isoT") const;		///< Gets the date/time in string format
	BError		setString(const BString dateTime);					///< Sets the date/time from string format

			// Operator functions
			operator BTime() const { return BTime(getSeconds()); }
	int		operator==(const BTimeUs& time) const { return (otime == time.otime); }
	int		operator!=(const BTimeUs& time) const { return (otime != time.otime); }
	int		operator>(const BTimeUs& time) const { return (otime > time.otime); }
	int		operator>=(const BTimeUs& time) const { return (otime >= time.otime); }
	int		operator<(const BTimeUs& time) const { return (otime < time.otime); }
	int		operator<=(const BTimeUs& time) const { return (otime <= time.otime); }
	BTimeUs		operator+(BInt64 microSeconds) const { return BTimeUs(otime + microSeconds); }
	BTimeUs&	operator+=(BInt64 microSeconds){ addMicroSeconds(microSeconds); return *this; }

private:
	BUInt64		otime;		///< Time in microseconds since 1970-01-01 to TAI standard
};

#endif 
