/*******************************************************************************
 *	BoapMc1.h	BoapMc1 RPC protocol
 *	T.Barnaby,	BEAM Ltd,	2018-06-21
 *	Copyright (c) 2018 All Right Reserved, Beam Ltd, http://www.beam.ltd.uk
 *******************************************************************************
 */
#ifndef	BoapMc1_h
#define	BoapMc1_h

#include <BTypes.h>
#include <BMutex.h>
#include <BSemaphore.h>
#include <BQueue.h>
#include <BFifo.h>
#include <BComms.h>

// Main BOAP Types
const BUInt16	BoapMc1Magic	= 0x5542;
enum BoapMc1Type		{ BoapMc1TypeRequest = 0x0000, BoapMc1TypeReply = 0x8000 };

// BoapMc1 packet header
struct BoapMc1PacketHead {
	BUInt16		magic;
	BUInt16		length;
	BUInt16		addressTo;			///< Do we want addresses ?
	BUInt16		addressFrom;
	BUInt16		cmd;
	BInt16		error;				///< Not used. Maybe return a BError object ?
	BUInt32		checksum;			///< Not used. Not sure we want this here or at the very end of the packet.
} __attribute__((aligned(8),packed));

// BoapMc1 packet
class BoapMc1Packet {
public:
	BoapMc1PacketHead	head;
	char			data[8];
} __attribute__((aligned(8),packed));

struct BoapMc1Error {
	BInt16		number;				///< The error number
	char		string[32];			///< The error string
} __attribute__((aligned(8),packed));

inline BUInt32 boapMc1CommsRoundupLen(BUInt32 len){
	return(len + 3) & ~3;
}

/*******************************************************************************
 *	Base for bidirectional communications objects
 *******************************************************************************
 */

class BoapMc1Comms {
public:
			BoapMc1Comms(Bool threaded = 0, BUInt reqSize = 512);
	virtual		~BoapMc1Comms();

	void		setCommsMode(Bool halfDuplex);				///< Sets half duplex mode
	void		setComms(BComms& comms);				///< Sets the communications interface to use
	void		setComms(BComms* comms);				///< Sets the communications interface to use
	void		setAddress(BUInt16 addressTo, BUInt16 addressFrom);	///< Sets the to and from addresses
	BUInt32		getApiVersion();					///< Returns the API version
	BUInt32		setTimeout(BUInt32 timeoutUs);				///< Sets the call timeout returning the current value

	BoapMc1Packet*	packetRx();						///< returns a reference to the current RX packet

	virtual BError	processRx(BTimeout timeoutUs = BTimeoutForever);	///< Process any RX packets queuing them as needed

protected:
	// Communications functions
	virtual BError		processRequests(BTimeout timeoutUs = BTimeoutForever);	///< Check and process any requests
	virtual BError		processRequest(BTimeout timeoutUs = BTimeoutForever);	///< Check and process any request

	BError			packetTx(BDataChunk* chunks, BUInt nChunks, BUInt16 waitCmdReply);

#ifdef ZAP_IDEAS
	BError			packetRxWait(BUInt num, char** data);		///< Wait for an amount of data and return pointer to the start of that data perhaps in a Fifo
	BError			packetRxWait(BUInt16 cmd);
#endif

	BError			packetRxData(void* data, BUInt nBytes);
	BError			packetRxEnd();

	Bool			othreaded;				///< Threaded operation
	BUInt32			oreqSize;				///< The maximum request size
	BMutex			olockCall;				///< Lock for RPC calls. Only one at a time
	BMutex			olockTx;				///< Lock for TX.
	BComms*			ocomms;
	BUInt32			oapiVersion;
	Bool			ohalfDuplex;				///< Half duplex mode
	BUInt32			otimeout;				///< The timeout in us for calls
	BUInt16			oaddressTo;
	BUInt16			oaddressFrom;

	BoapMc1Packet		opacketRxBase;
	BoapMc1Packet*		opacketRx;				///< The RX packet
	BoapMc1Packet		opacketTxBase;
	BoapMc1Packet*		opacketTx;				///< The TX packet
	BUInt			opacketRpcCmd;				///< Waiting for RPC reply to cmd
	BSemaphore		opacketRpcSema;				///< Wait RPC reply semaphore
	BSemaphore		opacketRpcDoneSema;			///< Wait RPC complete semaphore
	BoapMc1Error		oerror;					///< The call return error;
};


#endif
